@@ -63,7 +63,11 @@ static void Num_scalInv_Smatrix(SCALAR *x, PRIMME_INT m, int n, PRIMME_INT ldx,
 static int allocate_workspace_svds(primme_svds_params *primme_svds, int allocate);
 static int globalSum_Rprimme_svds(REAL *sendBuf, REAL *recvBuf, int count, 
       primme_svds_params *primme_svds);
-static void convTestFunAugmented(double *eval, void *evec, double *rNorm, int *isConv,
+static void compute_resNorm(SCALAR *leftsvec, SCALAR *rightsvec, REAL *rNorm,
+      primme_svds_params *primme_svds, int *ierr);
+static void default_convTestFun(double *sval, void *leftsvec, void *rightsvec,
+      double *rNorm, int *isConv, primme_svds_params *primme_svds, int *ierr);
+static void convTestFunAug(double *eval, void *evec, double *rNorm, int *isConv,
    primme_params *primme, int *ierr);
 static void convTestFunATA(double *eval, void *evec, double *rNorm, int *isConv,
    primme_params *primme, int *ierr);
@@ -149,6 +153,16 @@ int Sprimme_svds(REAL *svals, SCALAR *svecs, REAL *resNorms,
    int ret, allocatedTargetShifts;
    SCALAR *svecs0;
 
+   /* ------------------------------------ */
+   /* Set defaults for sequential programs */
+   /* ------------------------------------ */
+   if (primme_svds->numProcs <= 1 && svals != NULL && svecs != NULL && resNorms != NULL) {
+      primme_svds->mLocal = primme_svds->m;
+      primme_svds->nLocal = primme_svds->n;
+      primme_svds->procID = 0;
+      primme_svds->numProcs = 1;
+   }
+
    /* ------------------ */
    /* Set some defaults  */
    /* ------------------ */
@@ -181,9 +195,13 @@ int Sprimme_svds(REAL *svals, SCALAR *svecs, REAL *resNorms,
       return ALLOCATE_WORKSPACE_FAILURE;
    }
 
-   /* ----------------------- */
-   /* Set default monitor     */
-   /* ----------------------- */
+   /* ----------------------------------------- */
+   /* Set default monitor and convergence test  */
+   /* ----------------------------------------- */
+
+   if (!primme_svds->convTestFun) {
+      primme_svds->convTestFun = default_convTestFun;
+   }
 
    if (!primme_svds->monitorFun) {
       primme_svds->monitorFun = default_monitor;
@@ -273,7 +291,7 @@ static SCALAR* copy_last_params_from_svds(primme_svds_params *primme_svds, int s
    *allocatedTargetShifts = 0;
 
    if (method == primme_svds_op_none) {
-      primme->maxMatvecs = 1;
+      primme->maxMatvecs = 0;
       return NULL;
    }
 
@@ -302,13 +320,11 @@ static SCALAR* copy_last_params_from_svds(primme_svds_params *primme_svds, int s
 
    switch(method) {
    case primme_svds_op_AtA:
-      primme->convTestFun = convTestFunATA;
-      break;
    case primme_svds_op_AAt:
       primme->convTestFun = convTestFunATA;
       break;
    case primme_svds_op_augmented:
-      primme->convTestFun = convTestFunAugmented;
+      primme->convTestFun = convTestFunAug;
       break;
    case primme_svds_op_none:
       break;
@@ -357,7 +373,12 @@ static SCALAR* copy_last_params_from_svds(primme_svds_params *primme_svds, int s
    primme->iseed[1] = primme_svds->iseed[1];
    primme->iseed[2] = primme_svds->iseed[2];
    primme->iseed[3] = primme_svds->iseed[3];
-   primme->maxMatvecs = primme_svds->maxMatvecs;
+   if (stage == 0) {
+      primme->maxMatvecs = primme_svds->maxMatvecs/2;
+   }
+   else {
+      primme->maxMatvecs = primme_svds->maxMatvecs/2 - primme_svds->primme.stats.numMatvecs;
+   }
 
    primme->intWork = primme_svds->intWork;
    primme->intWorkSize = primme_svds->intWorkSize;
@@ -393,7 +414,7 @@ static SCALAR* copy_last_params_from_svds(primme_svds_params *primme_svds, int s
       } 
    }
    else if (stage == 1 && primme->targetShifts == NULL &&
-            primme_svds->target == primme_svds_smallest) {
+         primme_svds->target == primme_svds_smallest) {
 
       assert(method == primme_svds_op_augmented);
       *allocatedTargetShifts = 1;
@@ -637,7 +658,7 @@ int copy_last_params_to_svds(primme_svds_params *primme_svds, int stage,
    method = stage == 0 ? primme_svds->method : primme_svds->methodStage2;
 
    if (method == primme_svds_op_none) {
-      primme->maxMatvecs = 1;
+      primme->maxMatvecs = 0;
       return 0;
    }
 
@@ -754,7 +775,6 @@ int copy_last_params_to_svds(primme_svds_params *primme_svds, int stage,
    primme_svds->iseed[1] = primme->iseed[1];
    primme_svds->iseed[2] = primme->iseed[2];
    primme_svds->iseed[3] = primme->iseed[3];
-   primme_svds->maxMatvecs -= primme->stats.numMatvecs;
 
    /* Zero references to primme workspaces to prevent to be release by primme_free */
    primme->intWork = NULL;
@@ -765,9 +785,12 @@ int copy_last_params_to_svds(primme_svds_params *primme_svds, int stage,
       primme->targetShifts = NULL;
    }
 
-   /* Update residual norms when final stage */
-   if (primme_svds->methodStage2 != primme_svds_op_none) {
-      switch(method) {
+   /* Update residual norms. For normal equations we have to divide by the    */
+   /* the singular value. For the augmented, nothing is required because the  */
+   /* user defined stopping criterion computes the actual residual vector     */
+   /* norm and update the value.                                              */
+   
+   switch(method) {
       case primme_svds_op_AtA:
       case primme_svds_op_AAt:
          for (i=0; i<primme_svds->initSize; i++) {
@@ -778,10 +801,9 @@ int copy_last_params_to_svds(primme_svds_params *primme_svds, int stage,
          for (i=0; i<primme_svds->initSize; i++) {
             rnorms[i] *= sqrt(2.0);
          }
-         break;
+	 break;
       case primme_svds_op_none:
-         break;
-      }
+	 break;
    }
 
    return 0;
@@ -953,11 +975,149 @@ static int globalSum_Rprimme_svds(REAL *sendBuf, REAL *recvBuf, int count,
    return 0;
 }
 
+/*******************************************************************************
+ * Subroutine compute_resNorm - This routine computes the residual norm of a
+ *    given triplet (u,s,v):
+ *
+ *    sqrt(||A*v - s*u||^2 + ||A'*u - s*v||^2)
+ *
+ * NOTE:
+ *    - The given u and v may not have norm one.
+ *    - The computation requires two matvecs.
+ * 
+ * INPUT ARRAYS AND PARAMETERS
+ * ---------------------------
+ * leftsvec     The approximate left singular vector
+ * rightsvec    The approximate right singular vector
+ * primme_svds  Structure containing various solver parameters
+ *
+ * OUTPUT PARAMETERS
+ * ----------------------------------
+ * rNorm        The norm of the residual vector
+ * ierr         Error code
+ ******************************************************************************/
+
+static void compute_resNorm(SCALAR *leftsvec, SCALAR *rightsvec, REAL *rNorm,
+      primme_svds_params *primme_svds, int *ierr) {
+
+   int one = 1, notrans = 0, trans = 1;
+   PRIMME_INT nLocal = primme_svds->mLocal+primme_svds->nLocal;
+   SCALAR *Atu = (SCALAR*)malloc( sizeof(SCALAR)*nLocal);
+   if (Atu == NULL) {*ierr = 1; return;}
+   SCALAR *Av = &Atu[primme_svds->nLocal];
+
+   /* Av = A * v; Atu = A'u */
+
+   primme_svds->matrixMatvec(leftsvec, &primme_svds->mLocal, Atu,
+         &primme_svds->mLocal, &one, &trans, primme_svds, ierr);
+   if (*ierr != 0) {free(Atu); return;}
+   primme_svds->stats.numMatvecs++;
+   primme_svds->matrixMatvec(rightsvec, &primme_svds->nLocal, Av,
+         &primme_svds->nLocal, &one, &notrans, primme_svds, ierr);
+   if (*ierr != 0) {free(Atu); return;}
+   primme_svds->stats.numMatvecs++;
+
+   /* ip[0] = ||v|| */
+   /* ip[1] = ||u|| */
+   /* ip[2] = u'*A*v = u'*Av */
+
+   REAL ip0[3], ip[3];
+   ip0[0] = REAL_PART(Num_dot_Sprimme(primme_svds->nLocal, rightsvec, 1,
+            rightsvec, 1));
+   ip0[1] = REAL_PART(Num_dot_Sprimme(primme_svds->mLocal, leftsvec, 1,
+            leftsvec, 1));
+   ip0[2] = REAL_PART(Num_dot_Sprimme(primme_svds->mLocal, leftsvec, 1, Av, 1));
+   *ierr = globalSum_Rprimme_svds(ip0, ip, 3, primme_svds);
+   if (*ierr != 0) {free(Atu); return;}
+
+   ip[0] = sqrt(ip[0]);
+   ip[1] = sqrt(ip[1]);
+   REAL sval = ip[2]/ip[0]/ip[1];
+
+   /* If u'*A*v is negative, set rNorm as a large number */
+
+   if (sval < -0.0) {
+      *rNorm = HUGE_VAL;
+      free(Atu);
+      *ierr = 0;
+      return;
+   }
+
+   /* Atu = A'*u/||u|| - sval*v/||v|| */
+
+   Num_scal_Sprimme(primme_svds->nLocal, 1.0/ip[1], Atu, 1);
+   Num_axpy_Sprimme(primme_svds->nLocal, -sval/ip[0], rightsvec, 1, Atu, 1);
+
+   /* Av = A*v/||v|| - sval*u/||u|| */
+
+   Num_scal_Sprimme(primme_svds->mLocal, 1.0/ip[0], Av, 1);
+   Num_axpy_Sprimme(primme_svds->mLocal, -sval/ip[1], leftsvec, 1, Av, 1);
+
+   /* resNorm = sqrt(||A*v - s*u||^2 + ||A'*u - s*v||^2) = norm([Atu; Av]) */
+
+   REAL normr0;
+   normr0 = REAL_PART(Num_dot_Sprimme(nLocal, Atu, 1, Atu, 1));
+   *ierr = globalSum_Rprimme_svds(&normr0, rNorm, 1, primme_svds);
+   if (*ierr != 0) {free(Atu); return;}
+   *rNorm = sqrt(*rNorm);
+
+   free(Atu);
+   *ierr = 0;
+}
+
+/*******************************************************************************
+ * Subroutine default_convTestFun - This routine implements primme_params.
+ *    convTestFun and returns an approximate triplet converged when           
+ *    resNorm < eps * ||A||.
+ *
+ * INPUT ARRAYS AND PARAMETERS
+ * ---------------------------
+ * sval         The approximate singular value 
+ * leftsvec     The approximate left singular vector
+ * rightsvec    The approximate right singular vector
+ * rNorm        The norm of the residual vector
+ * primme_svds  Structure containing various solver parameters
+ *
+ * OUTPUT PARAMETERS
+ * ----------------------------------
+ * isConv      if it isn't zero the approximate pair is marked as converged
+ * ierr        error code
+ ******************************************************************************/
+
+static void default_convTestFun(double *sval, void *leftsvec_, void *rightsvec_,
+      double *rNorm, int *isConv, primme_svds_params *primme_svds, int *ierr) {
+
+   (void)sval; /* unused parameter */
+   const double machEps = MACHINE_EPSILON;
+   const double aNorm = primme_svds->aNorm;
+   SCALAR *leftsvec = (SCALAR*)leftsvec_, *rightsvec = (SCALAR*)rightsvec_;
+
+   *isConv = *rNorm < max(primme_svds->eps, machEps * 3.16) * aNorm;
+
+   /* If solving the augmented problem, the reported residual norm is an      */
+   /* approximation. Recheck the convergence criterion with the actual        */
+   /* residual norm when the convergence criterion is passed and the residual */
+   /* vector norm is from the augmented problem. When solving the augmented   */
+   /* problem the right and the left singular vectors are contiguous.         */
+
+   if (*isConv && leftsvec == rightsvec + primme_svds->nLocal && (
+            primme_svds->method == primme_svds_op_augmented
+            || primme_svds->methodStage2 == primme_svds_op_augmented)) {
+
+      REAL rnorm;
+      compute_resNorm(leftsvec, rightsvec, &rnorm, primme_svds, ierr);
+      if (*ierr != 0) return;
+
+      *isConv = rnorm < max(primme_svds->eps, machEps * 3.16) * aNorm;
+   }
+
+   *ierr = 0;
+}
+
 /*******************************************************************************
  * Subroutine convTestFunATA - This routine implements primme_params.
- *    convTestFun and returns an approximate eigenpair converged when           
- *    resNorm < eps * sval * primme_svds.aNorm = eps * sqrt(eval*primme.aNorm)
- *    resNorm is close to machineEpsilon * primme.aNorm.
+ *    convTestFun and calls primme_svds.convTestFun when solving normal
+ *    equations.
  *
  * INPUT ARRAYS AND PARAMETERS
  * ---------------------------
@@ -974,26 +1134,47 @@ static int globalSum_Rprimme_svds(REAL *sendBuf, REAL *recvBuf, int count,
 static void convTestFunATA(double *eval, void *evec, double *rNorm, int *isConv,
    primme_params *primme, int *ierr) {
 
-   const double machEps = MACHINE_EPSILON;
-   const double aNorm = (primme->aNorm > 0.0) ?
-      primme->aNorm : primme->stats.estimateLargestSVal;
-   (void)evec;  /* unused argument */
-   *isConv = *rNorm < max(
-               primme->eps * sqrt(fabs(*eval * aNorm)),
-               machEps * 3.16 * aNorm);
-   *ierr = 0;
+   primme_svds_params *primme_svds = (primme_svds_params *) primme->matrix;
+   primme_svds_operator method = &primme_svds->primme == primme ?
+      primme_svds->method : primme_svds->methodStage2;
+   assert(method == primme_svds_op_AtA || method == primme_svds_op_AAt);
+   double aNorm = (primme->aNorm > 0.0) ?
+            primme->aNorm : primme->stats.estimateLargestSVal;
+   double maxaNorm = max(primme->aNorm, primme->stats.estimateLargestSVal);
+
+   /* Check machine precision limit */
+
+   if (rNorm && *rNorm < MACHINE_EPSILON * maxaNorm * 3.16) {
+      *isConv = 1;
+      *ierr = 0;
+      return;
+   }
+
+   /* Update primme_svds->aNorm */
+
+   double oldaNorm = primme_svds->aNorm;
+   if (primme_svds->aNorm <= 0.0)
+      primme_svds->aNorm = sqrt(aNorm);
+
+   /* Call the callback */
+   
+   double sval = eval ? sqrt(fabs(*eval)) : 0.0;
+   double srNorm = (rNorm&&eval) ? *rNorm/sval : 0.0;
+   primme_svds->convTestFun(eval?&sval:NULL,
+      (method==primme_svds_op_AAt && evec) ? evec : NULL,
+      (method==primme_svds_op_AtA && evec) ? evec : NULL,
+      (rNorm&&eval)?&srNorm:NULL, isConv, primme_svds, ierr);
+
+   /* Restore aNorm */
+
+   primme_svds->aNorm = oldaNorm;
 }
 
+
 /*******************************************************************************
- * Subroutine convTestFunAugmented - This routine implements primme_params.
- *    convTestFun and returns an approximate eigenpair converged when           
- *
- *       sqrt(||Av - su||^2 + ||A'u - sv||^2) < ||A|| * eps = aNorm/sqrt(2)*eps
- *
- *    However, the previous test is expensive so it is only checked after the
- *    next one passes:
- *
- *       resNorm < ||A||*eps = aNorm/sqrt(2)*eps
+ * Subroutine convTestFunAug - This routine implements primme_params.
+ *    convTestFun and calls primme_svds.convTestFun when solving augmented
+ *    problem.
  *
  * INPUT ARRAYS AND PARAMETERS
  * ---------------------------
@@ -1007,89 +1188,38 @@ static void convTestFunATA(double *eval, void *evec, double *rNorm, int *isConv,
  * isConv      if it isn't zero the approximate pair is marked as converged
  ******************************************************************************/
 
-static void convTestFunAugmented(double *eval, void *evec_, double *rNorm,
-      int *isConv, primme_params *primme, int *ierr) {
+static void convTestFunAug(double *eval, void *evec, double *rNorm, int *isConv,
+   primme_params *primme, int *ierr) {
 
-   const double machEps = MACHINE_EPSILON;
-   const double aNorm = (primme->aNorm > 0.0) ?
-      primme->aNorm : primme->stats.estimateLargestSVal;
    primme_svds_params *primme_svds = (primme_svds_params *) primme->matrix;
-   SCALAR *evec = (SCALAR*)evec_;
+   assert(primme_svds_op_augmented == (&primme_svds->primme == primme ?
+            primme_svds->method : primme_svds->methodStage2));
+   double aNorm = (primme->aNorm > 0.0) ?
+            primme->aNorm : primme->stats.estimateLargestSVal;
 
-   /* Pre-test */
+   /* NOTE: Don't check machine precision limit of the residual norm.      */
+   /* Regardless of how small the residual is, we don't want to mark as    */
+   /* converged pairs that approximate the null space of the augmented     */
+   /* problem.                                                             */
 
-   *isConv = 
-      *rNorm < max(
-               primme->eps / sqrt(2.0) * aNorm,
-               machEps * 3.16 * aNorm) 
-      && *eval >= aNorm*machEps;
+   /* Update primme_svds->aNorm */
 
-   /* Actual test */
+   double oldaNorm = primme_svds->aNorm;
+   if (primme_svds->aNorm <= 0.0)
+      primme_svds->aNorm = aNorm;
 
-   if (*isConv && evec) {
-      int one = 1;
-      SCALAR *r = (SCALAR*)malloc(sizeof(SCALAR)*primme->nLocal);
-      if (r == NULL) {*ierr = 1; return;}
+   /* Call the callback */
 
-      /* r = [0 A';A 0] * evec = [ A'u; Av ] */
+   double sval = eval ? fabs(*eval) : 0.0;
+   double srNorm = rNorm ? *rNorm/sqrt(2.0) : 0.0;
+   primme_svds->convTestFun(eval?&sval:NULL,
+      evec?&((SCALAR*)evec)[primme_svds->nLocal]:NULL,
+      evec,
+      rNorm?&srNorm:NULL, isConv, primme_svds, ierr);
 
-      matrixMatvecSVDS(evec, &primme->nLocal, r, &primme->nLocal, &one, primme,
-            ierr);
-      if (*ierr != 0) return;
-      primme->stats.numMatvecs++;
-      
-      /* ip[0] = ||evec[0:nLocal-1]|| = ||v|| */
-      /* ip[1] = ||evec[nLocal:nLocal+mLocal-1]|| = ||u|| */
-      /* ip[2:4] = u'*A*v */
+   /* Restore aNorm */
 
-      REAL ip0[4], ip[4];
-      ip0[0] = REAL_PART(Num_dot_Sprimme(primme_svds->nLocal, evec, 1, evec,
-               1));
-      ip0[1] = REAL_PART(Num_dot_Sprimme(primme_svds->mLocal,
-               &evec[primme_svds->nLocal], 1, &evec[primme_svds->nLocal], 1));
-      ip0[3] = 0.0;
-      *(SCALAR*)&ip0[2] = Num_dot_Sprimme(primme_svds->mLocal,
-               &evec[primme_svds->nLocal], 1, &r[primme_svds->nLocal], 1);
-      *ierr = globalSum_Rprimme_svds(ip0, ip, 4, primme_svds);
-      if (*ierr != 0) return;
-      ip[0] = sqrt(ip[0]);
-      ip[1] = sqrt(ip[1]);
-      SCALAR sval = *(SCALAR*)&ip[2]/ip[0]/ip[1];
-      /* r[0:nLocal-1] = r[0:nLocal-1]/ip[1] - sval * evec[0:nLocal-1]/ip[0]  */
-      /*               = A'u/||u|| - sval*v/||v||                             */
-
-      Num_scal_Sprimme(primme_svds->nLocal, 1.0/ip[1], r, 1);
-      Num_axpy_Sprimme(primme_svds->nLocal, -sval/ip[0], evec, 1, r, 1);
-
-      /* r[nLocal:end] = r[nLocal:end]/ip[0] - sval * evec[nLocal:end]/ip[1] */
-      /*               = Av/||v|| - sval*u/||u||                             */
-
-      Num_scal_Sprimme(primme_svds->mLocal, 1.0/ip[0], &r[primme_svds->nLocal],
-            1);
-      Num_axpy_Sprimme(primme_svds->mLocal, -sval/ip[1],
-            &evec[primme_svds->nLocal], 1, &r[primme_svds->nLocal], 1);
-
-      /* normr = sqrt(||Av - su||^2 + ||A'u - sv||^2) */
-
-      REAL normr0, normr;
-      normr0 = REAL_PART(Num_dot_Sprimme(primme->nLocal, r, 1, r, 1));
-      *ierr = globalSum_Rprimme_svds(&normr0, &normr, 1, primme_svds);
-      if (*ierr != 0) return;
-      normr = sqrt(normr);
-
-      /* isConv = 1 iff normr <= ||A||*eps = aNorm/sqrt(2)*eps */
-
-      if (normr < max(aNorm/sqrt(2.0)*primme->eps, machEps * 5 * aNorm)) { 
-         *isConv = 1;
-      }
-      else {
-         *isConv = 0;
-      }
-
-      free(r);
-   }
-
-   *ierr = 0;
+   primme_svds->aNorm = oldaNorm;
 }
 
 
@@ -1346,7 +1476,7 @@ static void monitor_stage1(void *basisEvals_, int *basisSize, int *basisFlags,
    int numLocked0 = lockedEvals&&numLocked?*numLocked:0;
    int basisSize0 = (basisEvals&&basisSize?*basisSize:0) + numLocked0;
    REAL basisSvals[basisSize0], basisSVNorms[basisSize0];
-   int basisSVFlags[basisSize0], iblockSV[blockSize?*blockSize:1];
+   int basisSVFlags[basisSize0], iblockSV[blockSize&&*blockSize>0?*blockSize:1];
    int numConvergedSV = (numConverged?*numConverged:numLocked0);
 
    primme_svds_params *primme_svds = (primme_svds_params *) primme->matrix;
